"Mostly copied http://carnivore.it/2010/07/22/python_-_getifaddrs"
import ctypes
import socket
import sys


def getifaddrs():
    # AF_UNKNOWN / generic
    if sys.platform.startswith("darwin") or sys.platform.startswith("freebsd"):
        class sockaddr (ctypes.Structure):
            _fields_ = [
                ("sa_len", ctypes.c_uint8),
                ("sa_family", ctypes.c_uint8),
                ("sa_data", (ctypes.c_uint8 * 14))]
    else:
        class sockaddr(ctypes.Structure):
            _fields_ = [
                ("sa_family", ctypes.c_uint16),
                ("sa_data", (ctypes.c_uint8 * 14))
            ]

    # getifaddr structs
    class ifa_ifu_u(ctypes.Union):
        _fields_ = [
            ("ifu_broadaddr", ctypes.c_void_p),
            ("ifu_dstaddr", ctypes.c_void_p)
        ]

    class ifaddrs(ctypes.Structure):
        _fields_ = [
            ("ifa_next", ctypes.c_void_p),
            ("ifa_name", ctypes.c_char_p),
            ("ifa_flags", ctypes.c_uint),
            ("ifa_addr", ctypes.c_void_p),
            ("ifa_netmask", ctypes.c_void_p),
            ("ifa_ifu", ifa_ifu_u),
            ("ifa_data", ctypes.c_void_p)
        ]

    # AF_INET / IPv4
    class in_addr(ctypes.Union):
        _fields_ = [
            ("s_addr", ctypes.c_uint32),
        ]

    if sys.platform.startswith("darwin") or sys.platform.startswith("freebsd"):
        class sockaddr_in(ctypes.Structure):
            _fields_ = [
                ('sin_len', ctypes.c_uint8),
                ("sin_family", ctypes.c_uint8),
                ("sin_port", ctypes.c_uint16),
                ("sin_addr", ctypes.c_uint8 * 4),
                ("sin_zero", ctypes.c_uint8 * 8),  # padding
            ]
    else:
        class sockaddr_in(ctypes.Structure):
            _fields_ = [
                ("sin_family", ctypes.c_short),
                ("sin_port", ctypes.c_ushort),
                ("sin_addr", in_addr),
                ("sin_zero", (ctypes.c_char * 8)),  # padding
            ]

    # AF_INET6 / IPv6
    class in6_u(ctypes.Union):
        _fields_ = [
            ("u6_addr8", (ctypes.c_uint8 * 16)),
            ("u6_addr16", (ctypes.c_uint16 * 8)),
            ("u6_addr32", (ctypes.c_uint32 * 4))
        ]

    class in6_addr(ctypes.Union):
        _fields_ = [
            ("in6_u", in6_u),
        ]

    class sockaddr_in6(ctypes.Structure):
        _fields_ = [
            ("sin6_family", ctypes.c_short),
            ("sin6_port", ctypes.c_ushort),
            ("sin6_flowinfo", ctypes.c_uint32),
            ("sin6_addr", in6_addr),
            ("sin6_scope_id", ctypes.c_uint32),
        ]

    class sockaddr_ll(ctypes.Structure):
        _fields_ = [
            ("sll_family", ctypes.c_uint16),
            ("sll_protocol", ctypes.c_uint16),
            ("sll_ifindex", ctypes.c_uint32),
            ("sll_hatype", ctypes.c_uint16),
            ("sll_pktype", ctypes.c_uint8),
            ("sll_halen", ctypes.c_uint8),
            ("sll_addr", (ctypes.c_uint8 * 8))
        ]

    # AF_LINK / BSD|OSX
    class sockaddr_dl(ctypes.Structure):
        _fields_ = [
            ("sdl_len", ctypes.c_uint8),
            ("sdl_family", ctypes.c_uint8),
            ("sdl_index", ctypes.c_uint16),
            ("sdl_type", ctypes.c_uint8),
            ("sdl_nlen", ctypes.c_uint8),
            ("sdl_alen", ctypes.c_uint8),
            ("sdl_slen", ctypes.c_uint8),
            ("sdl_data", (ctypes.c_uint8 * 46))
        ]

    #libc = CDLL("libc.so.6")
    # Load library implementing getifaddrs and freeifaddrs.
    if sys.platform == 'darwin':
        libc = ctypes.CDLL('libc.dylib')
    else:
        libc = ctypes.CDLL('libc.so.6')

    ptr = ctypes.c_void_p(None)
    result = libc.getifaddrs(ctypes.pointer(ptr))
    if result:
        raise StopIteration()
    ifa = ifaddrs.from_address(ptr.value)

    while True:
        name = ifa.ifa_name.decode('utf-8')
        sa = sockaddr.from_address(ifa.ifa_addr)
        data = {}
        if sa.sa_family == socket.AF_INET:
            if ifa.ifa_addr is not None:
                si = sockaddr_in.from_address(ifa.ifa_addr)
                data['addr'] = socket.inet_ntop(socket.AF_INET, si.sin_addr)

                si = sockaddr_in.from_address(ifa.ifa_ifu.ifu_broadaddr)
                data['broadcast'] = socket.inet_ntop(
                    socket.AF_INET, si.sin_addr)
            if ifa.ifa_netmask is not None:
                si = sockaddr_in.from_address(ifa.ifa_netmask)
                data['netmask'] = socket.inet_ntop(
                    socket.AF_INET, si.sin_addr)

        if sa.sa_family == socket.AF_INET6:
            if ifa.ifa_addr is not None:
                si = sockaddr_in6.from_address(ifa.ifa_addr)
                data['addr'] = socket.inet_ntop(socket.AF_INET6, si.sin6_addr)
                if data['addr'].startswith('fe80:'):
                    data['scope'] = si.sin6_scope_id
            if ifa.ifa_netmask is not None:
                si = sockaddr_in6.from_address(ifa.ifa_netmask)
                data['netmask'] = socket.inet_ntop(
                    socket.AF_INET6, si.sin6_addr)

        if len(data) > 0:
            yield sa.sa_family, name, data


        if ifa.ifa_next:
            ifa = ifaddrs.from_address(ifa.ifa_next)
        else:
            break

    libc.freeifaddrs(ptr)


if __name__ == "__main__":
    from pprint import pprint
    pprint(getifaddrs())



import socket
import subprocess
import struct
import sys

from minifail import getifaddrs

def unpack_str_ip(ip):
    return struct.unpack("!L", socket.inet_pton(socket.AF_INET, ip))[0]


def make_network(addr, netmask):
    addr = unpack_str_ip(addr)
    netmask = unpack_str_ip(netmask)
    network = addr & netmask
    return network


def in_network(ip, network):
    ip = unpack_str_ip(ip)


def default_address(target_interface_name, target_ip, target_netmask):
    addresses = []

    for family, interface_name, data in getifaddrs.getifaddrs():
        if (family == socket.AF_INET and
            interface_name.startswith(target_interface_name)):
            addresses.append(data)

    target_network = make_network(target_ip, target_netmask)

    candidates = []
    for address in addresses:
        network = make_network(address['addr'], address['netmask'])
        if target_network == network:
            candidates.append(address)

    for address in candidates:
        if address['addr'] == target_ip:
            return address

    if len(candidates) > 0:
        return candidates[0]


def interface_with_ip(ip):
    for family, interface_name, data in getifaddrs.getifaddrs():
        if family == socket.AF_INET and data['addr'] == ip:
            return interface_name


def give_up_ip(ip):
    "XXX use ioctl"
    interface = interface_with_ip(ip)
    if sys.platform.startswith(("darwin", "freebsd")):
        command = ("ifconfig", interface, "delete", ip)
    else:
        command = ("ifconfig", interface, "down")

    subprocess.check_call(command)



import smtplib
from email.mime.text import MIMEText

class SMTPNotifier(object):
    def __init__(self, host, port, recipient, from_email):
        self.host = host
        self.port = port
        self.recipient = recipient
        self.from_email = from_email

    def notifiy(self, subject, body):
        msg = MIMEText(body)
        msg["Subject"] = subject
        msg["From"] = self.from_email
        msg["To"] = self.recipient

        connection = smtplib.SMTP(host=self.host, port=self.port)
        connection.sendmail(self.from_email, [self.recipient], msg.as_string())
        connection.quit()



"""Minifail, lower identifiers have the priority if master 1 and master 0 are
running, master 1 will give up the IP.

Usage:
  minifail.py [options] <identifier> <interface> <ip> <netmask>
  minifail.py -h | --help

Options:
  -h --help               Show this screen.
  --verbose               Be verbose
  --script=<script>       Script to execute when becoming master.
  --smtp_host=<host>      Port of the SMTP server for sending notifications (default localhost)
  --smtp_port=<port>      Port of the SMTP server for sending notifications (default 25)
  --recipient=<recipient> Recipient of email notifications (default root@localhost)
  --from=<from>           Sender of the email notifications (default root@localhost)
"""

import socket
import time
import sys
import subprocess

from docopt import docopt

from minifail import netutils
from minifail.notifier import SMTPNotifier

PORT = 1694
CHECK_PERIOD = 1 # in seconds
MAX_FAILURES = 3

notifier = None


def error(message, exit=True):
    global notifier

    sys.stderr.write("%s\n" % message)
    notifier.notifiy("Minifail error", message)
    if exit:
        sys.exit(1)


def add_ip(interface, ip, netmask, verbose=False):
    command = ["ifconfig", interface, "add", ip, "netmask", netmask, "up"]
    if verbose:
        print("Executing `%s`" % " ".join(command))
    subprocess.call(command)


class ConflictException(Exception):
    pass


def master_heartbeat(broadcast, identifier, verbose=False):
    broadcast_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    broadcast_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    broadcast_sock.setblocking(0)

    listen_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    listen_sock.setblocking(0)
    listen_sock.bind((broadcast, PORT))

    while True:
        while True:
            try:
                data, addr = listen_sock.recvfrom(16)
                try:
                    other_identifier = int(data)
                    if other_identifier < identifier:
                        raise ConflictException(
                                "Higher priority peer detected giving up")
                    if verbose:
                        print("Received heartbeat with priority %s" %
                                   other_identifier)
                except ValueError:
                    pass
            except socket.error:
                break

        sent = broadcast_sock.sendto(str(identifier), (broadcast, PORT))
        if sent != len(str(identifier)):
            error("Heartbeat sending failed")

        time.sleep(CHECK_PERIOD)


def become_master(interface, ip, netmask, script, verbose=False):
    global notifier

    add_ip(interface, ip, netmask, verbose=verbose)
    if script:
        subprocess.Popen([script])


def loop_until_master_not_beating(broadcast, verbose=False):
    broadcast_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    broadcast_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    broadcast_sock.setblocking(0)

    listen_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    listen_sock.setblocking(0)
    listen_sock.bind((broadcast, PORT))

    failures = 0

    while True:
        try:
            data, addr = listen_sock.recvfrom(16)
        except socket.error as e:
            if verbose:
                print("Missed a beat %r" % e)
            failures += 1
        else:
            failures = 0
            if verbose:
                print("received %s from %s" % (data, addr))

        if failures == MAX_FAILURES:
            return

        time.sleep(CHECK_PERIOD)


def main():
    global notifier
    arguments = docopt(__doc__, argv=sys.argv[1:], help=True, version=None)
    verbose = arguments["--verbose"]

    interface = arguments["<interface>"]
    target_ip = arguments["<ip>"]
    target_netmask = arguments["<netmask>"]
    identifier = int(arguments["<identifier>"])

    script = arguments.get("--script")

    host = arguments.get("--host", "localhost")
    port = arguments.get("--port", "25")
    recipient = arguments.get("--recipient", "root@localhost")
    from_email = arguments.get("--from", "root@localhost")

    notifier = SMTPNotifier(host, port, recipient, from_email)

    address = netutils.default_address(interface, target_ip, target_netmask)
    if not address:
        error("Couldn't find matching ip/interface")

    if address['addr'] != target_ip:
        if verbose:
            print("IP is not assigned to this machine, start monitoring")
        loop_until_master_not_beating(address['broadcast'], verbose=verbose)
        become_master(interface, target_ip, address['netmask'],
                      script, verbose=verbose)
        message = "minifail agent %d became master" % identifier
        notifier.notifiy(message, message)

    if verbose:
        print("Start broadcasting")

    try:
        master_heartbeat(address['broadcast'], identifier, verbose=verbose)
    except Exception as e:
        netutils.give_up_ip(target_ip)
        error_message = "%s: %s, IP has been given up" % (type(e), str(e))
        error(error_message, exit=False)

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(
    name='minifail',
    version='0.1.3',
    description='minifail is a simple IP failover compatible with Linux/BSD/MacOS',
    author='Gautier Hayoun',
    author_email='ghayoun@gmail.com',
    url='https://github.com/gautier/minifail',
    license='MIT',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'minifail = minifail.main:main',
        ]
    },
    install_requires=['docopt==0.5.0'],
)

